﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Problems.VehicleRouting.Variants;
using HeuristicLab.Random;

namespace HeuristicLab.Problems.VehicleRouting.Encodings.Potvin {
  [Item("GeographicDistanceClusterCreator", "Creates a VRP solution by clustering customers first with a KMeans-algorithm based on their geographic position and building tours afterwards alternatevly in a random or a greedy fashion.")]
  [StorableClass]
  public sealed class GeographicDistanceClusterCreator : ClusterCreator {

    [StorableConstructor]
    private GeographicDistanceClusterCreator(bool deserializing) : base(deserializing) { }

    public GeographicDistanceClusterCreator() : base() {
    }

    private GeographicDistanceClusterCreator(GeographicDistanceClusterCreator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GeographicDistanceClusterCreator(this, cloner);
    }


    public static PotvinEncoding CreateSolution(IVRPProblemInstance instance, IRandom random, int minK, int maxK, double clusterChangeThreshold, int creationOption) {
      PotvinEncoding result = new PotvinEncoding(instance);

      // (1) init data
      var coordinates = new List<double[]>(instance.Cities.Value);
      var pdp = instance as IPickupAndDeliveryProblemInstance;
      for (int i = 1; i <= instance.Cities.Value; i++) {
        if (pdp == null || pdp.GetDemand(i) >= 0)
          coordinates.Add(instance.GetCoordinates(i));
      }

      // (2) create a random number k of clusters
      int k = random.Next(minK, maxK);
      var kMeans = new KMeansAlgorithm<double[]>(CalculateMeanHelper, CalculateDistanceHelper);
      var clusters = kMeans.Run(coordinates, k, clusterChangeThreshold, random);

      // (3) build tours with a (a) shuffling (b) greedy tour creation routine
      foreach (var cluster in clusters) {
        Tour newTour = new Tour();
        result.Tours.Add(newTour);

        if (creationOption == 0) {
          // (a) shuffle
          cluster.ShuffleInPlace(random);
          foreach (var customer in cluster) {
            newTour.Stops.Add(customer + 1);
          }
        } else {
          // (b) greedy
          foreach (var customer in cluster) {
            newTour.Stops.Add(customer + 1);
          }
          GreedyTourCreation(instance, result, newTour, false);
        }
      }

      return result;
    }

    private static double[] CalculateMeanHelper(List<double[]> coordinates) {
      var mean = new double[coordinates[0].Length];
      foreach (double[] coord in coordinates) {
        for (int i = 0; i < mean.Length; i++) {
          mean[i] += coord[i] / coordinates.Count;
        }
      }
      return mean;
    }
    private static double CalculateDistanceHelper(double[] coord1, double[] coord2) {
      double distance = 0.0;
      for (int i = 0; i < coord1.Length; i++) {
        distance += Math.Pow(coord1[i] - coord2[i], 2);
      }
      return Math.Sqrt(distance);
    }

    public override IOperation InstrumentedApply() {
      IRandom random = RandomParameter.ActualValue;

      int minK = (MinK.Value.Value > 0) ? MinK.Value.Value : 1;
      int maxK = (MaxK.Value != null) ? MaxK.Value.Value : ProblemInstance.Vehicles.Value;
      double clusterChangeThreshold = (ClusterChangeThreshold.Value.Value >= 0.0 &&
                                       ClusterChangeThreshold.Value.Value <= 1.0)
        ? ClusterChangeThreshold.Value.Value
        : 0.0;

      // normalize probabilities
      double max = TourCreationProbabilities.Value.Max();
      double[] probabilites = new double[2];
      for (int i = 0; i < TourCreationProbabilities.Value.Length; i++) {
        probabilites[i] = TourCreationProbabilities.Value[i] / max;
      }

      List<int> creationOptions = new List<int>() { 0, 1 };
      int creationOption = creationOptions.SampleProportional(random, 1, probabilites, false, false).First();

      VRPToursParameter.ActualValue = CreateSolution(ProblemInstance, random, minK, maxK, clusterChangeThreshold, creationOption);
      return base.InstrumentedApply();
    }
  }
}
